#ifndef COBJ_H_2007
#define COBJ_H_2007
/*
 * Copyright (c) 2007, Andreas Stricker <andy@knitter.ch>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above 
 *     copyright notice, this list of conditions and the following 
 *     disclaimer in the documentation and/or other materials provided 
 *     with the distribution.
 *   * The names of its contributors must not be used to endorse or 
 *     promote products derived from this software without specific 
 *     prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Simple C Object model.
 *
 * This module is heavy inspired by the GLIB and cPython object models.
 *
 * The idea was to provide a simple an easy-to-use object library when
 * the glib is too heavy.
 * 
 * Of course the functionality is reduced to keep it simple.
 *
 * For an example to use have a look at cobj_test.c
 *
 * Each instantiable object must have the parent object as its first
 * member. This is due c guarantees that the first member always
 * starts at the same address. Therefore most manipulations are done
 * with casts.
 *
 * To make it clearer, the following graphic shows the structure:
 *
 * CObjBase
 *  +------- SubClass1 sc1
 *           +------------- SubSubclass1 ssc1
 *           +------------- SubSubclass2 ssc2
 *
 *                                                 +- - - - - - - - - - - -+
 * +- scc1 [SubSubClass]----+                      ¦                       ¦
 * |+- .base [SubClass] ---+|                      V                       ¦
 * ||+- .base [CObjBase] -+||    +- SubSubClass_type [SubSubClassType]-+   ¦
 * ||| .type - - - - - - -|||- ->|+- .parent [SubClassType] ----------+|   ¦
 * ||| .ref (counter)     |||    ||+- .parent [CObjType] ------------+||   ¦
 * ||+--------------------+||    ||| .typeid  - - - - - - - - - - - -|||- -+
 * || { SubClass members } ||    ||| .parent  - - - - - - - - - - - -|||- -+
 * |+----------------------+|    ||| .name = "SubSubClass"           |||   ¦
 * | { SubSubClass membs. } |    ||| { other type attributes }       |||   ¦
 * +------------------------+    ||+---------------------------------+||   ¦
 *                               || { SubClassType attributes }       ||   ¦
 *                               |+-----------------------------------+|   ¦
 *                               | { SubSubClassType attributes }      |   ¦
 *                               +-------------------------------------+   ¦
 *                                                                         ¦
 * +- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
 * ¦                 +- - - - - - - - - - -+
 * ¦                 ¦                     ¦
 * ¦                 V                     ¦               +- - - - - - - - +
 * ¦  +- SubClass_type [SubClassType] -+   ¦               ¦                ¦
 * +->|+- .parent [CObjType] ---------+|   ¦               V                ¦
 *    || .typeid - - - - - - - - - - -||- -+    +- CObjType_type [CObjType]-+ ¦
 *    || .parent - - - - - - - - - - -||- - - ->| .typeid  - - - - - - - - -|-+
 *    || .name = "SubClass"           ||        | .parent = NULL            |
 *    || { other type attributes }    ||        | .name = "CObjBase"        |
 *    |+------------------------------+|        | { other type attributes } |
 *    | { SubClassType attributes }    |        +---------------------------+
 *    +--------------------------------+
 *
 * @version 0.1
 */                                                                          

#include <unistd.h> /* size_t */

/** 
 * Version of this library.
 * 0xAAIIBB (AA = Major, II = Minor, BB = Build)
 */
#define COBJ_VERSION 0x000100

/** Convert an instance to base type instance */
#define COBJ(self_) \
    ((CObjBase*)(self_))

/** Convert a type to base type */
#define COBJTYPE(type_) \
    ((CObjType*)(type_))

/** Return the type of an instance */
#define COBJ_GET_TYPE(obj_) \
    ((obj_) ? ((CObjBase*)(obj_))->type : NULL)

/** Check if the instance is a type of subclass of a type */
#define COBJ_CHECK_TYPE(obj_, type_) \
    cobjtype_check_type(COBJ_GET_TYPE(obj_), COBJTYPE(type_))

/** Instantiate a new instance of given type */
#define cobj_new(type_) \
    ((type_ *)cobjbase_new(COBJTYPE(&type_ ## _type)))

/** increment reference counter for an instance */
#define cobj_ref(object_) \
    cobjbase_ref(COBJ(object_))

/** 
 * decrement reference counter for an instance 
 * an call destructor and free resources if reference
 * counter reach zero 
 */
#define cobj_unref(object_) \
    cobjbase_unref(COBJ(object_))

typedef struct CObjType CObjType;   /**< Base type */
typedef struct CObjBase CObjBase;   /**< Base instance type */

typedef int (*cobjfunc_init)(CObjBase *self);   /**< initializer */
typedef int (*cobjfunc_free)(CObjBase *self);   /**< deallocator */

/** Base type descriptor */
struct CObjType {
    CObjType *typeid;           /**< type of this class */
    CObjType *parent;           /**< type of parent class */
    size_t object_size;         /**< size of the object */
    size_t type_size;           /**< size of this class */
    const char *name;           /**< name of this class */

    cobjfunc_init initialize;   /**< initializer for members */
    cobjfunc_free deallocate;   /**< deallocator function to free members */
};

/** Base instance descriptor (Head of all objects) */
struct CObjBase {
    CObjType *type;             /**< pointer to type object */
    int ref;                    /**< reference counter */
};

CObjType CObjBase_type;         /**< base type object */

/**
 * Initialize the cobj library (mainly to initialize base type)
 */
void cobjtype_initialize();

/**
 * Register a new type
 */
CObjType* cobjtype_register(CObjType *type);

/**
 * Check if type is the same type or a subclass of checktype
 */
int cobjtype_check_type(CObjType *type, CObjType *checktype);

/**
 * Instantiate a new object of specified type
 */
CObjBase* cobjbase_new(CObjType *type);

/**
 * Decrement reference counter for object self and 
 * call destructor and deallocator if the reference
 * counter reach zero
 */
void cobjbase_unref(CObjBase *self);

/**
 * Increment reference counter for object
 */
CObjBase* cobjbase_ref(CObjBase *self);

/**
 * Free an object independend of reference counter.
 * This must not be used if reference counting is
 * used and you don't exactly know what you do.
 *
 * This may be usefull to break a reference counting
 * loop up.
 */
int cobjbase_free(CObjBase *self);

#endif /* COBJ_H_2007 */

