#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>

#include "cobj.h"

/* MyObject section */

#define MYOBJECT(obj_) \
    ((MyObject*)(obj_))
#define MYOBJECTTYPE(type_) \
    ((MyObjectType*)(type_))

typedef struct MyObject MyObject;
typedef struct MyObjectType MyObjectType;

static int myobject__i__deallocate(MyObject *self);
static void myobject__i__set_text(MyObject *self, const char *text);

struct MyObject {
    CObjBase base;

    size_t length;
    char *text;
};

struct MyObjectType {
    CObjType parent;

    void (*set_text)(MyObject *self, const char *text);
};

MyObjectType MyObject_type = {
    .parent = {
        .typeid         = (CObjType*)&MyObject_type,
        .parent         = (CObjType*)&CObjBase_type,
        .object_size    = sizeof(MyObject),
        .type_size      = sizeof(MyObjectType),
        .name           = "MyObject",

        .initialize     = NULL,
        .deallocate     = (cobjfunc_free)myobject__i__deallocate,
    },
    .set_text = myobject__i__set_text,
};

static int
myobject__i__deallocate(MyObject *self)
{
    assert(COBJ_CHECK_TYPE(self, &MyObject_type));
    if (self->text)
        free(self->text);
    return 1;
}

static void
myobject__i__set_text(MyObject *self, const char *text)
{
    assert(COBJ_CHECK_TYPE(self, &MyObject_type));
    if (self->text)
        free(self->text);
    self->length = strlen(text);
    self->text = strdup(text);
}

void
myobject_set_text(MyObject *self, const char *text)
{
    assert(COBJ_CHECK_TYPE(self, &MyObject_type));
    assert(MYOBJECTTYPE(self->base.type)->set_text);
    MYOBJECTTYPE(self->base.type)->set_text(self, text);
}

/* MySubObject */

#define MYSUBOBJECT(obj_) \
    ((MySubObject*)(obj_))
#define MYSUBOBJECTTYPE(type_) \
    ((MySubObjectType*)(type_))

typedef struct MySubObject MySubObject;
typedef struct MySubObjectType MySubObjectType;

static void mysubobject__i__set_text(MyObject *self, const char *text);
static void mysubobject__i__set_toupper(MySubObject *self, int value);

struct MySubObject {
    MyObject base;

    int toupper;
};

struct MySubObjectType {
    MyObjectType parent;

    void (*set_toupper)(MySubObject *self, int value);
};

MySubObjectType MySubObject_type = {
    .parent = {
        .parent = {
            .typeid         = (CObjType*)&MySubObject_type,
            .parent         = (CObjType*)&MyObject_type,
            .object_size    = sizeof(MySubObject),
            .type_size      = sizeof(MySubObjectType),
            .name           = "MySubObject",

            .initialize     = NULL,
            .deallocate     = (cobjfunc_free)myobject__i__deallocate,
        },
        .set_text = mysubobject__i__set_text,
    },
    .set_toupper = mysubobject__i__set_toupper,
};

static void
mysubobject__i__set_text(MyObject *self, const char *text)
{
    char *ch;
    MyObject_type.set_text(self, text);

    assert(COBJ_CHECK_TYPE(self, &MySubObject_type));

    printf("toupper = %d\n", MYSUBOBJECT(self)->toupper);
    if (MYSUBOBJECT(self)->toupper)
        for (ch = self->text; *ch != 0; ch++)
            *ch = toupper(*ch);
}

static void
mysubobject__i__set_toupper(MySubObject *self, int value)
{
    assert(COBJ_CHECK_TYPE(self, &MySubObject_type));

    self->toupper = value;
}

void
mysubobject_set_toupper(MySubObject *self, int value)
{
    assert(COBJ_CHECK_TYPE(self, &MySubObject_type));
    assert(MYSUBOBJECTTYPE(COBJ_GET_TYPE(self))->set_toupper);
    MYSUBOBJECTTYPE(COBJ_GET_TYPE(self))->set_toupper(self, value);
}

static CObjBase* keep_obj;

int main()
{
    CObjBase *base;
    MyObject *obj;
    MySubObject *subobj;

    cobjtype_initialize();
    cobjtype_register(COBJTYPE(&MyObject_type));
    cobjtype_register(COBJTYPE(&MySubObject_type));

    base = cobj_new(CObjBase);
    obj = cobj_new(MyObject);
    subobj = cobj_new(MySubObject);

    printf("object type %p name: %s\n", COBJ_GET_TYPE(base), COBJ(base)->type->name);
    printf("object type %p name: %s\n", COBJ_GET_TYPE(obj), COBJ(obj)->type->name);
    printf("object type %p name: %s\n", COBJ_GET_TYPE(subobj), COBJ(subobj)->type->name);

    myobject_set_text(obj, "Hello World");

    printf("text[%d]: %s\n", obj->length, obj->text);

    mysubobject_set_toupper(subobj, 0);
    myobject_set_text(MYOBJECT(subobj), "Yo! Hello");

    printf("text[%d] no upper: %s\n", 
            MYOBJECT(subobj)->length, MYOBJECT(subobj)->text);

    mysubobject_set_toupper(subobj, 1);
    myobject_set_text(MYOBJECT(subobj), "Yo! Hello");

    printf("text[%d] with upper: %s\n", 
            MYOBJECT(subobj)->length, MYOBJECT(subobj)->text);

    keep_obj = cobj_ref(obj);

    printf("ref# for %s: %d\n", COBJ_GET_TYPE(base)->name, COBJ(base)->ref);
    printf("ref# for %s: %d\n", COBJ_GET_TYPE(obj)->name, COBJ(obj)->ref);
    printf("ref# for %s: %d\n", COBJ_GET_TYPE(subobj)->name, COBJ(subobj)->ref);

    cobj_unref(base);
    cobj_unref(obj);
    cobj_unref(subobj);

    printf("ref# for %s: %d\n", COBJ_GET_TYPE(obj)->name, COBJ(obj)->ref);

    cobj_unref(keep_obj);

    return 0;
}

